unit tgl;

interface

uses
  windows;

type
  TOnGlInit = procedure;
  TOnGlClose = procedure;
  TOnGlPaint = procedure;
  TOnGlResize = procedure(Width, height: Integer);

function  gl_open(title: string; width, height: Integer): Integer;
procedure gl_close;
procedure set_onglinit(NewOnGlInit: TOnGlInit);
procedure set_onglpaint(NewOnGlPaint: TOnGlPaint);
procedure set_onglclose(NewOnGlClose: TOnGlClose);
procedure set_onglresize(NewOnGlResize: TOnGlResize);

var
  wnd: HWND;
  window_hdc: HDC;

implementation
{$ifdef __TGL_ICON__}
{$R 'tinyptc.res'}
{$endif}

uses
  opengl, Messages;

const
  __PTC_ICON__: string = 'IDI_MAIN';
  // menu option identifier
  SC_ZOOM_MSK: Integer = $400;
  SC_ZOOM_1x1: Integer = $401;
  SC_ZOOM_2x2: Integer = $402;
  SC_ZOOM_4x4: Integer = $404;

type
  HDRAWDIB  = THandle;
  PVOID     = Pointer;
  TRCOptions = set of (opDoubleBuffered, opGDI, opStereo);

var
  wc: WNDCLASS;
  onglinit:  TOnGlInit;
  onglclose: TOnGlClose;
  onglPaint: TOnGlPaint;
  onglResize: TOnGlResize;
  RC: HGLRC;

function  CreateRenderingContext(DC: HDC; Options: TRCOptions; ColorBits, StencilBits, AccumBits, AuxBuffers: Integer;
  Layer: Integer): HGLRC;
const
  MemoryDCs = [OBJ_MEMDC, OBJ_METADC, OBJ_ENHMETADC];
var
  PFDescriptor: TPixelFormatDescriptor;
  PixelFormat: Integer;
  AType: DWORD;
begin
  FillChar(PFDescriptor, SizeOf(PFDescriptor), 0);
  with PFDescriptor do begin
    nSize := SizeOf(PFDescriptor);
    nVersion := 1;
    dwFlags := PFD_SUPPORT_OPENGL;
    AType := GetObjectType(DC);
    if AType in MemoryDCs then dwFlags := dwFlags or PFD_DRAW_TO_BITMAP
                          else dwFlags := dwFlags or PFD_DRAW_TO_WINDOW;
    if opDoubleBuffered in Options then dwFlags := dwFlags or PFD_DOUBLEBUFFER;
    if opGDI in Options then dwFlags := dwFlags or PFD_SUPPORT_GDI;
    if opStereo in Options then dwFlags := dwFlags or PFD_STEREO;
    iPixelType := PFD_TYPE_RGBA;
    cColorBits := ColorBits;
    cDepthBits := 32;
    cStencilBits := StencilBits;
    cAccumBits := AccumBits;
    cAuxBuffers := AuxBuffers;
    if Layer = 0 then iLayerType := PFD_MAIN_PLANE
                 else
      if Layer > 0 then iLayerType := PFD_OVERLAY_PLANE
                   else iLayerType := Byte(-1) {PFD_UNDERLAY_PLANE, only Inprise knows why they have declared
                                                an obvious Byte value as LongWord};
  end;

  // just in case it didn't happen already
  PixelFormat := ChoosePixelFormat(DC, @PFDescriptor);
  // NOTE: It is not allowed to change a pixel format of a device context once it has been set.
  //       So cache DC and RC always somewhere to avoid re-setting the pixel format!
  SetPixelFormat(DC, PixelFormat, @PFDescriptor);

  // check the properties just set
  DescribePixelFormat(DC, PixelFormat, SizeOf(PFDescriptor), PFDescriptor);

  Result := wglCreateLayerContext(DC, Layer);
end;

function WndProc(HWindow: HWnd; Message, WParam, LParam: Longint): Longint; stdcall;
{$ifdef __TGL_RESIZE_WINDOW__}
var
  windowsize: TRECT;
{$endif}
var
  painthdc: HDC;
  ps: PaintStruct;
begin
  Result := 0;
	// handle message
	case (message) of
		WM_PAINT:
      begin
        painthdc := BeginPaint(HWindow, ps);

        wglMakeCurrent(painthdc, RC);

        if assigned(onglpaint) then
          onglpaint;

        SwapBuffers(painthdc);
        wglMakeCurrent(0, 0);

        EndPaint(HWindow, ps);

{
        wglMakeCurrent(window_hdc, RC);
        if assigned(onglpaint) then
          onglpaint;

        SwapBuffers(window_hdc);
        wglMakeCurrent(0, 0);

        // validate window
        ValidateRect(wnd, nil);
}
      end;
{$ifdef __TGL_RESIZE_WINDOW__}
		WM_SIZE:
      begin
        GetClientRect(HWindow, windowsize);

        wglMakeCurrent(window_hdc, RC);
        if assigned(onglResize) then
          onglResize(windowsize.right, windowsize.bottom);
        wglMakeCurrent(0,0);
      end;
{$endif}
{$ifdef __TGL_CLOSE_ON_ESCAPE__}
    WM_KEYDOWN:
      // close on escape key
      if ((wParam and $FF) = 27) then
        SendMessage(HWindow, WM_CLOSE, 0, 0);
{$endif}
{$ifdef __TGL_ALLOW_CLOSE__}
    WM_CLOSE:
      begin
        // close tgl
		    gl_close();

{$ifndef __TGL_WINDOWED__}
        ChangeDisplaySettings(devmode(nil^), 0);
        ShowCursor(True);
{$endif}

	      // exit process
		    ExitProcess(0);
      end;
{$endif}
  	else
  		// unhandled messages
			Result := DefWindowProc(HWindow, message, wParam, lParam);
  end;
end;

function gl_open(title: string; width, height: Integer): Integer;
var
{$ifdef __TGL_WINDOWED__}
  rect: TRect;
  original_window_x: Integer;
  original_window_y: Integer;
{$else}
  dmScreenSettings: DEVMODE;
{$endif}
  message: MSG;
begin
  result :=  0;

{$ifdef __TGL_WINDOWED__}
	// register window class
	wc.style := CS_OWNDC or CS_VREDRAW or CS_HREDRAW;      // very important
	wc.lpfnWndProc := @WndProc;
	wc.cbClsExtra := 0;
	wc.cbWndExtra := 0;
{$ifdef __TGL_ICON__}
	wc.hInstance := GetModuleHandle(nil);
	wc.hIcon := LoadIcon(wc.hInstance, PChar(__PTC_ICON__));
{$else}
	wc.hInstance := 0;
	wc.hIcon := 0;
{$endif}
	wc.hCursor := LoadCursor(0, IDC_ARROW);
	wc.hbrBackground := 0;
	wc.lpszMenuName := nil;
	wc.lpszClassName := PChar(Title);
	if Windows.RegisterClass(wc) = 0 then
    Exit;

	// calculate window size
	rect.left   := 0;
	rect.top    := 0;
	rect.right  := width;
	rect.bottom := height;
{$ifdef __TGL_RESIZE_WINDOW__}
	AdjustWindowRect(rect, WS_OVERLAPPEDWINDOW, False);
{$else}
	AdjustWindowRect(rect, WS_POPUP or WS_SYSMENU or WS_CAPTION, False);
{$endif}
	rect.right  := rect.right - rect.left;
	rect.bottom := rect.bottom - rect.top;

{$ifdef __TGL_CENTER_WINDOW__}
	// center window
	original_window_x := (GetSystemMetrics(SM_CXSCREEN) - rect.right) shr 1;
	original_window_y := (GetSystemMetrics(SM_CYSCREEN) - rect.bottom) shr 1;
{$else}
	// let windows decide
	original_window_x := Integer(CW_USEDEFAULT);
	original_window_y := Integer(CW_USEDEFAULT);
{$endif}

	// create window and show it
{$ifdef __TGL_RESIZE_WINDOW__}
	wnd := CreateWindowEx(0, PChar(title), PChar(title), WS_OVERLAPPEDWINDOW,
                         original_window_x, original_window_y, rect.right,
                          rect.bottom, 0, 0, 0, nil);

{$else}
	wnd := CreateWindowEx(0, PChar(title), PChar(title),
                         WS_OVERLAPPEDWINDOW and (not WS_MAXIMIZEBOX) and (not WS_THICKFRAME),
                          original_window_x, original_window_y, rect.right,
                           rect.bottom, 0, 0, 0, nil);
{$endif}

	ShowWindow(wnd, SW_NORMAL);

{$else} // ifdef windowed
	// register window class
	wc.style := CS_OWNDC;
	wc.lpfnWndProc := @WndProc;
	wc.cbClsExtra := 0;
	wc.cbWndExtra := 0;
{$ifdef __TGL_ICON__}
	wc.hInstance := GetModuleHandle(nil);
	wc.hIcon := LoadIcon(wc.hInstance, PChar(__PTC_ICON__));
{$else}
	wc.hInstance := 0;
	wc.hIcon := 0;
{$endif}
	wc.hCursor := LoadCursor(0, IDC_ARROW);
	wc.hbrBackground := 0;
	wc.lpszMenuName := nil;
	wc.lpszClassName := PChar(Title);
	if Windows.RegisterClass(wc) = 0 then
    Exit;

    // create window
{$ifdef __TGL_ICON__}
  wnd := CreateWindow(PChar(title), PChar(title), WS_POPUP or WS_SYSMENU,
                        0, 0, Width, Height, 0, 0, 0, nil);
{$else}
  wnd := CreateWindow(PChar(title), PChar(title),WS_POPUP,
                        0, 0, Width, Height, 0, 0, 0, nil);
{$endif}

	ShowWindow(wnd, SW_NORMAL);

  ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));

  // Set parameters for the screen setting
  with dmScreenSettings do begin
    dmSize       := SizeOf(dmScreenSettings);
    dmPelsWidth  := width;              // Window width
    dmPelsHeight := Height;             // Window height
    dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT;
  end;

  // Try to change screen mode to fullscreen
  if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    exit;

  ShowCursor(False);                // Turn off the cursor
{$endif}

	// get window dc
	window_hdc := GetDC(wnd);
  RC := CreateRenderingContext(window_hdc, [opDoubleBuffered], 16, 0, 0, 0, 0);

  wglMakeCurrent(window_hdc, RC);

  if assigned(onglinit) then
    onglinit;

  wglMakeCurrent(0, 0);

{$ifdef __TGL_DISABLE_SCREENSAVER__}
    // disable screensaver while tgl is open
    SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 0, nil, 0);
{$endif}

  wglMakeCurrent(window_hdc, RC);
  if assigned(onglResize) then
     onglResize(width, height);
  wglMakeCurrent(0,0);

  while true do begin
    // invalidate window
    InvalidateRect(wnd, nil, TRUE);

    // send paint window message
    PostMessage(wnd, WM_PAINT, 0, 0);

    // process messages
    while (PeekMessage(message, wnd, 0, 0, PM_REMOVE)) do begin
      // translate and dispatch
      TranslateMessage(message);
      DispatchMessage(message);
    end;

    // sleep
    Sleep(0);
  end;

	// success
	result := 1;
end;

procedure gl_close;
begin
  if assigned(onglclose) then
    onglclose;

  wglMakeCurrent(0, 0);
  if RC <> 0 then
    wglDeleteContext(RC);

	// release DC
	ReleaseDC(wnd, window_hdc);

	// destroy window
	DestroyWindow(wnd);

{$ifdef __TGL_DISABLE_SCREENSAVER__}
  // enable screensaver now that tgl is closed
  SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 1, nil, 0);
{$endif}
end;

procedure set_onglinit(NewOnGlInit: TOnGlInit);
begin
  onglinit := NewOnGlInit;
end;

procedure set_onglpaint(NewOnGlPaint: TOnGlPaint);
begin
  onglPaint := NewOnGlPaint;
end;

procedure set_onglclose(NewOnGlClose: TOnGlClose);
begin
  onglclose := NewOnGlClose;
end;

procedure set_onglresize(NewOnGlResize: TOnGlResize);
begin
  onglresize := NewOnGlResize;
end;

initialization
  onglinit   := nil;
  onglclose  := nil;
  onglPaint  := nil;
  onglResize := nil;
end.
