unit uswirl;

// __TGL_RESIZE_WINDOW__        Resizable window
// __TGL_ALLOW_CLOSE__          window can be closed
// __TGL_CLOSE_ON_ESCAPE__      Close window when escape is pressed
// __TGL_WINDOWED__             Window or fullscreen
// __TGL_CENTER_WINDOW__        Place window in center of screen
// __TGL_DISABLE_SCREENSAVER__  Disable the screensaver
// __TGL_ICON__                 Show Icon

interface

procedure MainFunc;


implementation

uses
  mmsystem,
  windows,
  messages,
  opengl,
  tgl;

const
  SWIRLSIZE = 50;
  CILSIZE = 10;
// 50, 10 for draw 3
// 100, 1 for draw 4

type
  TCilArray = array[0..CILSIZE,0..2] of double;
  PByteA = ^TByteA;
  TByteA = array [0..0] of Byte;

function ArcTan2(Y, X: Extended): Extended;
asm
        FLD     Y
        FLD     X
        FPATAN
        FWAIT
end;

function  FixedgluBuild2DMipmaps(target: GLenum; components, width, height: GLint;
            format, atype: GLenum; data: Pointer): Integer; stdcall; external 'glu32.dll' name 'gluBuild2DMipmaps';


var
  WIDTH: Integer;
  HEIGHT: Integer;
  swirl:         array [0..7, 0..SWIRLSIZE, 0..2] of double;
  sat, cat:      array [0..CILSIZE] of double;
  pal:           array [0..SWIRLSIZE,0..2] of byte;
{$ifdef __TGL_WINDOWED__}
  NrFrame:   Integer;
  startTime: double;
{$endif}

function FloatToStr(Value: Extended): string;
begin
  Result := '';
  if Value > 1000 then
    Result := Result + Char(Ord('0') + Trunc(Value/1000));
  if Value > 100 then begin
    Value := value - 1000 * Trunc(Value/1000);
    Result := Result + Char(Ord('0') + Trunc(Value/100));
  end;
  if Value > 10 then begin
    Value := value - 100 * Trunc(Value/100);
    Result := Result + Char(Ord('0') + Trunc(Value/10));
  end;
  Value := value - 10 * Trunc(Value/10);
  Result := Result + Char(Ord('0') + Trunc(Value)) + '.';
  Value := value - Trunc(Value);
  Value := Value * 10;
  Result := Result + Char(Ord('0') + Trunc(Value));
  Value := value - Trunc(Value);
  Value := Value * 10;
  Result := Result + Char(Ord('0') + Trunc(Value));
end;

procedure bol2cart(var x,y,z: double; const a,h,zm: double);
var
  th: double;
begin
	th := cos(h) * zm;
	x  := cos(a) * th;
	y  := sin(a) * th;
	z  := sin(h) * zm;
end;

procedure cart2bol(var a,h,zm: double; const x,y,z: double);
var
  th: double;
begin
  zm := sqrt(x*x + y*y + z*z);
	if ((y = 0.0) and (x = 0.0)) then
		a := 0
	else
	  a := arctan2(y, x);
	th := sqrt(x*x + y*y);
	if ((th = 0.0) and  (z = 0.0)) then
		h := 0
	else
	  h := arctan2(z, th);
end;

procedure Calcswirl;
var
 rx, ry,rz, lag, t, ti: double;
 n, i: integer;
 pb, point: array[0..7,0..2] of double;
 tx,ty,s: double;
 sr,cr: double;
begin
  ti := 0.5 * timegettime * 0.001;
  t := ti + 0.05 * (cos(ti * 0.41 + 0.123) + cos(ti * 0.51 + 0.234) +
                     cos(ti * 0.61 + 0.623) + cos(ti * 0.21 + 0.723));
  lag := 0.7 * 0.005 * (cos(ti * 0.41 + 0.0123) + cos(ti * 0.51 + 0.0234) +
                         cos(ti * 0.61 + 0.0623) + cos(ti * 0.021 + 1.723));

  point[0][0] := -1;
  point[0][1] := 1;
  point[0][2] := 1;
  point[1][0] := 1;
  point[1][1] := 1;
  point[1][2] := 1;
  point[2][0] := 1;
  point[2][1] := -1;
  point[2][2] := 1;
  point[3][0] := -1;
  point[3][1] := -1;
  point[3][2] := 1;
  point[4][0] := -1;
  point[4][1] := 1;
  point[4][2] := -1;
  point[5][0] := 1;
  point[5][1] := 1;
  point[5][2] := -1;
  point[6][0] := 1;
  point[6][1] := -1;
  point[6][2] := -1;
  point[7][0] := -1;
  point[7][1] := -1;
  point[7][2] := -1;
  move(point, pb, 8 * 3 * sizeOf(double));

  s:= 0;
  for i := 0 to SWIRLSIZE do begin
    move(pb, point, 8 * 3 * sizeOf(double));
    for n := 0 to 7 do begin
      rx := 2.4 * t;
      ry := 2.5 * t;
      rz := 2.6 * t;

      tx := point[n][0];
      ty := point[n][1];
      sr := sin(rx);
      cr := cos(rx);
      point[n][0] := cr * tx - sr * ty;
      point[n][1] := sr * tx + cr * ty;

      tx := point[n][0];
      ty := point[n][2];
      sr := sin(ry);
      cr := cos(ry);
      point[n][0] := cr * tx - sr * ty;
      point[n][2] := sr * tx + cr * ty;

      tx := point[n][1];
      ty := point[n][2];
      sr := sin(rz);
      cr := cos(rz);
      point[n][1] := cr * tx - sr * ty;
      point[n][2] := sr * tx + cr * ty;

      point[n][0] := point[n][0] * s;
      point[n][1] := point[n][1] * s;
      point[n][2] := point[n][2] * s;

      move(point[n], swirl[n, i], 3 * sizeof(double));
    end;

    s := s + 0.6/SWIRLSIZE;
    t := t - lag;
  end;
end;

procedure Drawswirl1;
var
 i, j: integer;
begin
  glcolor3ub(255,255,155);
  glbegin(GL_POINTS);
  for i := 0 to SWIRLSIZE do begin
    for j := 0 to 7 do begin
      glVertex3d(swirl[j][i][0], swirl[j][i][1], swirl[j][i][2]);
    end;
  end;

  glend();
end;

procedure Drawswirl2;
var
 i, j: integer;
 pal: array [0..SWIRLSIZE,0..2] of byte;
begin
  for i := 0 to SWIRLSIZE do begin
    pal[SWIRLSIZE - i][0] := (i * 255) div SWIRLSIZE;
    pal[SWIRLSIZE - i][1] := (i * 255) div SWIRLSIZE;
    pal[SWIRLSIZE - i][2] := round(sqrt(i/SWIRLSIZE) * 255);
  end;
  for j := 0 to 7 do begin
    glbegin(GL_LINE_STRIP);
    for i := 0 to SWIRLSIZE do begin
      glcolor3ub(pal[i][0], pal[i][1], pal[i][2]);
      glVertex3d(swirl[j][i][0], swirl[j][i][1], swirl[j][i][2]);
    end;
    glend();
  end;
end;

procedure SetAngle(var cilpos: TCilArray; const dx,dy,dz: double);
var
  x,y,z,x1,y1,z1: double;
  a,h,zm: double;
  i: integer;
begin
  cart2bol(a, h, zm, dx, dy, dz);
	bol2cart(x,  y,  z,  a + 0.5 * PI, 0,            1);
	bol2cart(x1, y1, z1, a ,           h + 0.5 * PI, 1);

  for i := 0 to CILSIZE - 1 do begin
    cilpos[i,0] := sat[i] * x + cat[i] * x1;
    cilpos[i,1] := sat[i] * y + cat[i] * y1;
    cilpos[i,2] := sat[i] * z + cat[i] * z1;
  end;
end;


procedure SetCilpos(i,j: integer; var Cilpos: TCilArray);
var
  ppos, npos, dir: array [0..2] of double;
begin
  if (i = 0) then begin
    move(swirl[i,0] ,ppos , 3 * sizeof(double));
    move(swirl[i,1] ,npos , 3 * sizeof(double));
  end else if (i = SWIRLSIZE) then begin
    move(swirl[i, SWIRLSIZE - 1] ,ppos , 3 * sizeof(double));
    move(swirl[i, SWIRLSIZE] ,npos , 3 * sizeof(double));
  end else begin
    move(swirl[i, j - 1] ,ppos , 3 * sizeof(double));
    move(swirl[i, j + 1] ,npos , 3 * sizeof(double));
  end;

  dir[0] := npos[0] - ppos[0];
  dir[1] := npos[1] - ppos[1];
  dir[2] := npos[2] - ppos[2];

  if ((dir[0] = 0) and (dir[1] = 0) and (dir[2] = 0)) then begin
    dir[0] := 0;
    dir[1] := 0;
    dir[2] := 1;
  end;

  SetAngle(cilpos, dir[0], dir[1], dir[2]);

	cilpos[CILSIZE][0] := cilpos[0][0];
  cilpos[CILSIZE][1] := cilpos[0][1];
  cilpos[CILSIZE][2] := cilpos[0][2];
end;

procedure Drawswirl3;
var
 i, j, k: integer;
 cilpos, ocilpos, ncil, oncil: TCilArray;
 tubesize: double;
begin
  glEnable(GL_LIGHT1);
  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);

  glcolor3ub(255,255,155);
  glBegin(GL_Quads);
  for i := 0 to 7 do begin
    setcilpos(i, 0, cilpos);
    move(cilpos, ncil, (CILSIZE + 1) * 3 * sizeof(double));
    tubesize := 0.05;

    for k := 0 to CILSIZE do begin
      cilpos[k][0] := cilpos[k][0] * tubesize + swirl[i, 0][0];
      cilpos[k][1] := cilpos[k][1] * tubesize + swirl[i, 0][1];
      cilpos[k][2] := cilpos[k][2] * tubesize + swirl[i, 0][2];
    end;

    for j := 1 to SWIRLSIZE  do begin
      tubesize := 0.05 * (1 - j/SWIRLSIZE);
      move(cilpos, ocilpos, (CILSIZE + 1) * 3 * sizeof(double));
      move(ncil, oncil, (CILSIZE + 1) * 3 * sizeof(double));
      setcilpos(i, j, cilpos);
      move(cilpos, ncil, (CILSIZE + 1) * 3 * sizeof(double));

      for k := 0 to CILSIZE do begin
        cilpos[k][0] := cilpos[k][0] * tubesize + swirl[i, j][0];
        cilpos[k][1] := cilpos[k][1] * tubesize + swirl[i, j][1];
        cilpos[k][2] := cilpos[k][2] * tubesize + swirl[i, j][2];
      end;

      for k := 0 to CILSIZE -1 do begin
         glcolor3ub(pal[j][0], pal[j][1], pal[j][2]);
         glNormal3d(ncil[k][0], ncil[k][1], ncil[k][2]);
         glVertex3d(cilpos[k][0], cilpos[k][1], cilpos[k][2]);

         glNormal3d(ncil[k+1][0], ncil[k+1][1], ncil[k+1][2]);
         glVertex3d(cilpos[k+1][0], cilpos[k+1][1], cilpos[k+1][2]);

         glcolor3ub(pal[j - 1][0], pal[j - 1][1], pal[j - 1][2]);
         glNormal3d(oncil[k+1][0], oncil[k+1][1], oncil[k+1][2]);
         glVertex3d(ocilpos[k+1][0], ocilpos[k+1][1], ocilpos[k+1][2]);

         glNormal3d(oncil[k][0], oncil[k][1], oncil[k][2]);
         glVertex3d(ocilpos[k][0], ocilpos[k][1], ocilpos[k][2]);
      end;
    end;
  end;

  glend();
end;

procedure Drawswirl4;
var
 i, j: integer;
 fb_buffer: array[0 .. SWIRLSIZE * 8 * 4] of GLFloat;
 x,y,w: double;
begin
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);

  glClear(GL_COLOR_BUFFER_BIT);

  glFeedbackBuffer(SWIRLSIZE * 8 * 4, GL_3D, Pointer(@fb_buffer[0]));
  glRenderMode(GL_FEEDBACK);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity;
  gluLookAt( 4.5, 0, 0,
             0, 0, 0,
             0, 0, 1);

  glbegin(GL_POINTS);
  for i := 0 to 7 do begin
    for j := 0 to SWIRLSIZE - 1 do begin
      glVertex3d(swirl[i][j][0], swirl[i][j][1], swirl[i][j][2]);
    end;
  end;

  glend();
  glPopMatrix();
  glRenderMode(GL_RENDER);

  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glColor3f(0.5, 0.35, 0.20);

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glScalef(2.0 / width, 2.0 / Height, 1.0);
  glTranslatef(0.5 * -width, 0.5 * -Height, 0.0);

  glEnable(GL_TEXTURE_2D);
  glBlendFunc(GL_ONE, GL_ONE);
  glEnable(GL_BLEND);

  glBegin(GL_QUADS);

  for i := 0 to 8 * (SWIRLSIZE) - 1 do begin
    x := fb_buffer[4 * i + 1];
    y := fb_buffer[4 * i + 2];

    w := 13 - 10 * (i mod SWIRLSIZE)/ SWIRLSIZE;

    glTexCoord2f(0.0, 0.0);
    glVertex2f(x - w, y - w);

    glTexCoord2f(1.0, 0.0);
    glVertex2f(x + w, y - w);

    glTexCoord2f(1.0, 1.0);
    glVertex2f(x + w, y + w);

    glTexCoord2f(0.0, 1.0);
    glVertex2f(x - w, y + w);
  end;

  glEnd();

  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
end;

procedure ResizeGL(Width, height: Integer);
begin
  glViewport(0 , 0, Width, Height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
  gluPerspective(30, Width/Height, 3, 6);
end;

procedure InitGL;
const
  glfLightDiffuse : Array[0..3] of GLfloat = (0.7, 0.7, 0.7, 1.0);
  glfLightPosition: Array[0..3] of GLfloat = (100.0, 100.0, 1000.0, 1.0);
var
  i,j: integer;
  x,y,r,d: double;
  texbuf: array[0..4095] of byte;
begin
  glEnable(GL_DEPTH_TEST);

  glLightfv(GL_LIGHT1, GL_DIFFUSE , @glfLightDiffuse);
  glLightfv(GL_LIGHT1, GL_POSITION, @glfLightPosition);

  glColorMaterial(GL_FRONT, GL_DIFFUSE);
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  for i := 0 to SWIRLSIZE do begin
    pal[SWIRLSIZE - i][0] := (i * 191) div SWIRLSIZE + 64;
    pal[SWIRLSIZE - i][1] := (i * 191) div SWIRLSIZE + 64;
    pal[SWIRLSIZE - i][2] := 64 + round(sqrt(i/SWIRLSIZE) * 127);
  end;

  for i := 0 to CILSIZE - 1 do begin
    sat[i] := sin((2 * PI * i)/CILSIZE);
    cat[i] := cos((2 * PI * i)/CILSIZE);
  end;

  r := 32;
  for i := 0 to 63 do begin
    for j := 0 to 63 do begin
      x := abs(i - 32);
      y := abs(j - 32);
      d := sqrt(x * x + y * y);
      if d < r then begin
        d := 1 - (d/r);
        texbuf[i * 64 + j] := Round(255 * d * d);
      end else begin
        texbuf[i * 64 + j] := 0;
      end;
    end;
  end;

  FixedgluBuild2DMipmaps(GL_TEXTURE_2D, 1, 64, 64,
		    GL_Luminance,
		    GL_UNSIGNED_BYTE, PbyteA(@texbuf[0]));

  glTexParameteri(GL_TEXTURE_2D,
		  GL_TEXTURE_MIN_FILTER,
		  GL_LINEAR_MIPMAP_NEAREST);
  glTexParameteri(GL_TEXTURE_2D,
		  GL_TEXTURE_MAG_FILTER,
		  GL_LINEAR);

end;

procedure PaintGL;
{$ifdef __TGL_WINDOWED__}
var
  s: string;
  t: double;
{$endif}
begin
  glClearColor(0.1, 0.1, 0.3, 0);

  glClear(GL_DEPTH_BUFFER_BIT or GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity;
  gluLookAt( 4.5, 0, 0,
             0, 0, 0,
             0, 0, 1);

  CalcSwirl;
  DrawSwirl3;

{$ifdef __TGL_WINDOWED__}
  Inc(NrFrame);
  t := (timeGetTime * 0.001) - StartTime;
  if t > 2 then begin
    s := 'test ' + FloatToStr(NrFrame/t) + ' fps';

    SendMessage(wnd, WM_SETTEXT, 0, Longint(Pchar(s)));

    startTime := timeGetTime * 0.001;
    NrFrame := 0;
  end;
{$endif}
end;

procedure MainFunc;
begin
{$ifdef __TGL_WINDOWED__}
  startTime := timeGetTime * 0.001;
  NrFrame := 0;
{$endif}

  Width := 640;
  height := 480;

  set_onglinit(InitGl);
  set_onglpaint(PaintGl);
  set_onglresize(ResizeGl);

  gl_open('test', WIDTH, HEIGHT);
end;

end.

